/*
 * Copyright (c) 2006 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 * Copyright (c) 2006 The Technical University of Chemnitz. All 
 *                    rights reserved.
 *
 * Author(s): (alphabetically)
 *            Sven Eckelmann
 *            Torsten Hoefler 
 *            Mirko Reinhard 
 *            Matthias Treydte 
 */


/*
 * Copyright (c) 2004-2007 The Trustees of Indiana University and Indiana
 *                         University Research and Technology
 *                         Corporation.  All rights reserved.
 * Copyright (c) 2004-2006 The University of Tennessee and The University
 *                         of Tennessee Research Foundation.  All rights
 *                         reserved.
 * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2004-2005 The Regents of the University of California.
 *                         All rights reserved.
 * $COPYRIGHT$
 * 
 * Additional copyrights may follow
 * 
 * $HEADER$
 */
/**
 * @file
 */
#ifndef MCA_BTL_ETH_H
#define MCA_BTL_ETH_H

#include "ompi_config.h"
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

/* Open MPI includes */
#include "opal/event/event.h"
#include "ompi/class/ompi_free_list.h"
#include "ompi/mca/btl/btl.h"
#include "ompi/mca/btl/base/base.h"
#include "ompi/mca/mpool/mpool.h" 
#include "ompi/mca/btl/btl.h"
#include "opal/class/opal_hash_table.h"

#include <netinet/ether.h>


#define MCA_BTL_ETH_STATISTICS 0
BEGIN_C_DECLS


/**
 * ETH BTL component.
 */

struct mca_btl_eth_component_t {
    mca_btl_base_component_2_0_0_t super;   /**< base BTL component */
    uint32_t eth_addr_count;                /**< total number of addresses */
    uint32_t eth_num_btls;                  /**< number of hcas available to the ETH component */
    uint32_t eth_num_links;                 /**< number of logical links per physical device */
    struct mca_btl_eth_module_t **eth_btls; /**< array of available BTL modules */
    struct mca_btl_eth_proc_t* eth_local;   /**< local proc struct */
    int eth_free_list_num;                  /**< initial size of free lists */
    int eth_free_list_max;                  /**< maximum size of free lists */
    int eth_free_list_inc;                  /**< number of elements to alloc when growing free lists */
    int eth_endpoint_cache;                 /**< amount of cache on each endpoint */
    opal_hash_table_t eth_procs;            /**< hash table of eth proc structures */
    opal_list_t eth_events;                 /**< list of pending eth events */
    opal_mutex_t eth_lock;                  /**< lock for accessing module state */

    opal_event_t eth_recv_event;            /**< recv event for listen socket */
    int eth_listen_sd;                      /**< listen socket for incoming connection requests */
    unsigned short eth_listen_port;         /**< listen port */
    int32_t eth_port_min;                   /**< minimum port */
    int32_t eth_port_range;                 /**< port range */
    /* Port range restriction */

    char*  eth_if_include;                  /**< comma seperated list of interface to include */
    char*  eth_if_exclude;                  /**< comma seperated list of interface to exclude */
    int    eth_sndbuf;                      /**< socket sndbuf size */
    int    eth_rcvbuf;                      /**< socket rcvbuf size */
    int    eth_echo_attempts;               /**< amount of tries to reach a remote proc via the bound interface */
    int    eth_echo_timeout;                /**< timeout of each echo request */


    /* free list of fragment descriptors */
    ompi_free_list_t eth_frag_eager;
    ompi_free_list_t eth_frag_max;
    ompi_free_list_t eth_frag_user;
}; 
typedef struct mca_btl_eth_component_t mca_btl_eth_component_t;

OMPI_MODULE_DECLSPEC extern mca_btl_eth_component_t mca_btl_eth_component;

/**
 * BTL Module Interface
 */
struct mca_btl_eth_module_t {
    mca_btl_base_module_t  super;  /**< base BTL interface */
    mca_btl_base_recv_reg_t eth_reg[256];
    uint16_t           eth_ifkindex; /** <BTL kernel interface index */
    struct ether_addr  eth_ifaddr;  /**< BTL interface address */
    opal_list_t        eth_endpoints;
#if MCA_BTL_ETH_STATISTICS
    size_t eth_bytes_sent;
    size_t eth_bytes_recv;
    size_t eth_send_handler;
#endif
}; 
typedef struct mca_btl_eth_module_t mca_btl_eth_module_t;
extern mca_btl_eth_module_t mca_btl_eth_module;

#if defined(__WINDOWS__)
#define CLOSE_THE_SOCKET(socket)   closesocket(socket)
#else
#define CLOSE_THE_SOCKET(socket)   close(socket)
#endif  /* defined(__WINDOWS__) */

/**
 * Register ETH component parameters with the MCA framework
 */
extern int mca_btl_eth_component_open(void);

/**
 * Any final cleanup before being unloaded.
 */
extern int mca_btl_eth_component_close(void);

/**
 * ETH component initialization.
 * 
 * @param num_btl_modules (OUT)           Number of BTLs returned in BTL array.
 * @param allow_multi_user_threads (OUT)  Flag indicating wether BTL supports user threads (TRUE)
 * @param have_hidden_threads (OUT)       Flag indicating wether BTL uses threads (TRUE)
 */
extern mca_btl_base_module_t** mca_btl_eth_component_init(
    int *num_btl_modules, 
    bool allow_multi_user_threads,
    bool have_hidden_threads
);

/**
 * ETH component control.
 */
int mca_btl_eth_component_control(
    int param, 
    void* value, 
    size_t size
);


/**
 * ETH component progress.
 */
extern int mca_btl_eth_component_progress(void);



/**
 * Cleanup any resources held by the BTL.
 * 
 * @param btl  BTL instance.
 * @return     OMPI_SUCCESS or error status on failure.
 */

extern int mca_btl_eth_finalize(
    struct mca_btl_base_module_t* btl
);


/**
 * PML->BTL notification of change in the process list.
 * 
 * @param btl (IN)
 * @param nprocs (IN)     Number of processes
 * @param procs (IN)      Set of processes
 * @param peers (OUT)     Set of (optional) peer addressing info.
 * @param peers (IN/OUT)  Set of processes that are reachable via this BTL.
 * @return     OMPI_SUCCESS or error status on failure.
 * 
 */

extern int mca_btl_eth_add_procs(
    struct mca_btl_base_module_t* btl,
    size_t nprocs,
    struct ompi_proc_t **procs,
    struct mca_btl_base_endpoint_t** peers,
    opal_bitmap_t* reachable
);

/**
 * PML->BTL notification of change in the process list.
 *
 * @param btl (IN)     BTL instance
 * @param nproc (IN)   Number of processes.
 * @param procs (IN)   Set of processes.
 * @param peers (IN)   Set of peer data structures.
 * @return             Status indicating if cleanup was successful
 *
 */

extern int mca_btl_eth_del_procs(
    struct mca_btl_base_module_t* btl,
    size_t nprocs,
    struct ompi_proc_t **procs,
    struct mca_btl_base_endpoint_t** peers
);


/**
 * Initiate an asynchronous send.
 *
 * @param btl (IN)         BTL module
 * @param endpoint (IN)    BTL addressing information
 * @param descriptor (IN)  Description of the data to be transfered
 * @param tag (IN)         The tag value used to notify the peer.
 */

extern int mca_btl_eth_send(
    struct mca_btl_base_module_t* btl,
    struct mca_btl_base_endpoint_t* btl_peer,
    struct mca_btl_base_descriptor_t* descriptor,
    mca_btl_base_tag_t tag
);


/**
 * Initiate an asynchronous put.
 *
 * @param btl (IN)         BTL module
 * @param endpoint (IN)    BTL addressing information
 * @param descriptor (IN)  Description of the data to be transferred
 */

extern int mca_btl_eth_put(
    struct mca_btl_base_module_t* btl,
    struct mca_btl_base_endpoint_t* btl_peer,
    struct mca_btl_base_descriptor_t* decriptor
);


/**
 * Initiate an asynchronous get.
 *
 * @param btl (IN)         BTL module
 * @param endpoint (IN)    BTL addressing information
 * @param descriptor (IN)  Description of the data to be transferred
 */

extern int mca_btl_eth_get(
    struct mca_btl_base_module_t* btl,
    struct mca_btl_base_endpoint_t* btl_peer,
    struct mca_btl_base_descriptor_t* decriptor
);

/**
 * Register a callback function that is called on receipt
 * of a fragment.
 *
 * @param btl (IN)     BTL module
 * @return             Status indicating if registration was successful
 *
 */

extern int mca_btl_eth_register(
    struct mca_btl_base_module_t* btl, 
    mca_btl_base_tag_t tag, 
    mca_btl_base_module_recv_cb_fn_t cbfunc, 
    void* cbdata); 
    
/**
 * Allocate a descriptor with a segment of the requested size.
 * Note that the BTL layer may choose to return a smaller size
 * if it cannot support the request.
 *
 * @param btl (IN)      BTL module
 * @param size (IN)     Request segment size.
 */

extern mca_btl_base_descriptor_t* mca_btl_eth_alloc(
    struct mca_btl_base_module_t* btl,
    struct mca_btl_base_endpoint_t* endpoint,
    uint8_t order,
    size_t size,
    uint32_t flags); 


/**
 * Return a segment allocated by this BTL.
 *
 * @param btl (IN)      BTL module
 * @param descriptor (IN)  Allocated descriptor.
 */

extern int mca_btl_eth_free(
    struct mca_btl_base_module_t* btl, 
    mca_btl_base_descriptor_t* des); 
    

/**
 * Prepare a descriptor for send/rdma using the supplied
 * convertor. If the convertor references data that is contigous,
 * the descriptor may simply point to the user buffer. Otherwise,
 * this routine is responsible for allocating buffer space and
 * packing if required.
 *
 * @param btl (IN)          BTL module
 * @param endpoint (IN)     BTL peer addressing
 * @param convertor (IN)    Data type convertor
 * @param reserve (IN)      Additional bytes requested by upper layer to precede user data
 * @param size (IN/OUT)     Number of bytes to prepare (IN), number of bytes actually prepared (OUT) 
*/

mca_btl_base_descriptor_t* mca_btl_eth_prepare_src(
    struct mca_btl_base_module_t* btl,
    struct mca_btl_base_endpoint_t* peer,
    struct mca_mpool_base_registration_t*,
    struct opal_convertor_t* convertor,
    uint8_t order,
    size_t reserve,
    size_t* size,
    uint32_t flags
);

extern mca_btl_base_descriptor_t* mca_btl_eth_prepare_dst( 
    struct mca_btl_base_module_t* btl, 
    struct mca_btl_base_endpoint_t* peer,
    struct mca_mpool_base_registration_t*,
    struct opal_convertor_t* convertor,
    uint8_t order,
    size_t reserve,
    size_t* size,
    uint32_t flags); 


#define ETH_ARR2ADDR(arr)    \
    (*((struct ether_addr*)arr))


#define ETH_ADDR2ARR(addr)    \
    ((addr).ether_addr_octet)


static __inline__ unsigned long long rdtsc_(void)
{
  unsigned long long int x;
     __asm__ volatile (".byte 0x0f, 0x31" : "=A" (x));
     return x;
}

static __inline void rdtsc_old(unsigned long long *x)
{
  /* cpuid - guarantees that previous instructions are completed */
  /* rdtsc - Read Time-Stamp Counter (64-bit) into EDX:EAX */
  __asm__ __volatile__(
       "xorl %%eax, %%eax\n\t"
       "cpuid\n\t"
       "rdtsc\n\t"
       "movl %0, %%ebx\n\t"
       "movl %%eax, (%%ebx)\n\t"
       "movl %%edx, 4(%%ebx)\n\t"
       "xorl %%eax, %%eax\n\t"
       "cpuid"
       : 
       : "m" (x)
       : "edx", "eax", "ebx", "ecx"
  );
}  
/**
  * Fault Tolerance Event Notification Function
  * @param state Checkpoint Stae
  * @return OMPI_SUCCESS or failure status
  */
int mca_btl_eth_ft_event(int state);

END_C_DECLS
#endif
