/*
 * Copyright (c) 2006 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 * Copyright (c) 2006 The Technical University of Chemnitz. All 
 *                    rights reserved.
 *
 * Author(s): (alphabetically)
 *            Sven Eckelmann
 *            Torsten Hoefler 
 *            Mirko Reinhard 
 *            Matthias Treydte 
 */

/*
 * Copyright (c) 2004-2007 The Trustees of Indiana University and Indiana
 *                         University Research and Technology
 *                         Corporation.  All rights reserved.
 * Copyright (c) 2004-2007 The University of Tennessee and The University
 *                         of Tennessee Research Foundation.  All rights
 *                         reserved.
 * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2004-2005 The Regents of the University of California.
 *                         All rights reserved.
 * Copyright (c) 2007      Cisco, Inc.  All rights reserved.
 * $COPYRIGHT$
 * 
 * Additional copyrights may follow
 * 
 * $HEADER$
 *
 */

#include "ompi_config.h"

#include "opal/opal_socket_errno.h"
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <string.h>
#include <fcntl.h>
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#include <sys/ioctl.h>		/* ioctl constants */
#include <net/if.h>		/* struct ifreq {for ioctl()} */
#include <netinet/ether.h>

#include "ompi/constants.h"
#include "opal/event/event.h"
#include "opal/util/if.h"
#include "opal/util/output.h"
#include "opal/util/argv.h"
#include "orte/types.h"
#include "ompi/mca/btl/btl.h"

#include "opal/mca/base/mca_base_param.h"
#include "ompi/runtime/ompi_module_exchange.h"
#include "ompi/mca/mpool/base/base.h" 
#include "ompi/mca/btl/base/btl_base_error.h"
#include "btl_eth.h"
#include "btl_eth_addr.h"
#include "btl_eth_proc.h"
#include "btl_eth_frag.h"
#include "btl_eth_endpoint.h" 
#include "ompi/mca/btl/base/base.h" 


mca_btl_eth_component_t mca_btl_eth_component = {
    {
        /* First, the mca_base_component_t struct containing meta information
           about the component itself */

        {
            MCA_BTL_BASE_VERSION_2_0_0,

            "eth", /* MCA component name */
            OMPI_MAJOR_VERSION,  /* MCA component major version */
            OMPI_MINOR_VERSION,  /* MCA component minor version */
            OMPI_RELEASE_VERSION,  /* MCA component release version */
            mca_btl_eth_component_open,  /* component open */
            mca_btl_eth_component_close  /* component close */
        },
        {
            /* The component is checkpoint ready */
            MCA_BASE_METADATA_PARAM_CHECKPOINT
        },

        mca_btl_eth_component_init,  
        NULL,
    }
};

/*
 * utility routines for parameter registration
 */

static inline char* mca_btl_eth_param_register_string(
        const char* param_name, 
        const char* help_string,
        const char* default_value)
{
    char *value;
    mca_base_param_reg_string(&mca_btl_eth_component.super.btl_version,
                              param_name, help_string, false, false,
                              default_value, &value);
    return value;
}

static inline int mca_btl_eth_param_register_int(
        const char* param_name, 
        const char* help_string,
        int default_value)
{
    int value;
    mca_base_param_reg_int(&mca_btl_eth_component.super.btl_version,
                           param_name, help_string, false, false, 
                           default_value, &value);
    return value;
}


/*
 * Data structure for accepting connections.
 */

struct mca_btl_eth_event_t {
    opal_list_item_t item;
    opal_event_t event;
};
typedef struct mca_btl_eth_event_t mca_btl_eth_event_t;

static void mca_btl_eth_event_construct(mca_btl_eth_event_t* event)
{
    OPAL_THREAD_LOCK(&mca_btl_eth_component.eth_lock);
    opal_list_append(&mca_btl_eth_component.eth_events, &event->item);
    OPAL_THREAD_UNLOCK(&mca_btl_eth_component.eth_lock);
}

static void mca_btl_eth_event_destruct(mca_btl_eth_event_t* event)
{
    OPAL_THREAD_LOCK(&mca_btl_eth_component.eth_lock);
    opal_list_remove_item(&mca_btl_eth_component.eth_events, &event->item);
    OPAL_THREAD_UNLOCK(&mca_btl_eth_component.eth_lock);
}

OBJ_CLASS_INSTANCE(
    mca_btl_eth_event_t,
    opal_list_item_t,
    mca_btl_eth_event_construct,
    mca_btl_eth_event_destruct);


/*
 * functions for receiving event callbacks
 */
static void mca_btl_eth_component_recv_handler(int, short, void*);
static void mca_btl_eth_component_accept_handler(int, short, void*);


/*
 *  Called by MCA framework to open the component, registers
 *  component parameters.
 */

int mca_btl_eth_component_open(void)
{
#ifdef __WINDOWS__
    WSADATA win_sock_data;
    if( WSAStartup(MAKEWORD(2,2), &win_sock_data) != 0 ) {
        BTL_ERROR(("failed to initialise windows sockets:%d", WSAGetLastError()));
        return OMPI_ERROR;
    }
#endif

    /* initialize state */
    mca_btl_eth_component.eth_listen_sd = -1;
    mca_btl_eth_component.eth_num_btls=0;
    mca_btl_eth_component.eth_addr_count = 0;
    mca_btl_eth_component.eth_btls=NULL;
    
    /* initialize objects */ 
    OBJ_CONSTRUCT(&mca_btl_eth_component.eth_lock, opal_mutex_t);
    OBJ_CONSTRUCT(&mca_btl_eth_component.eth_procs, opal_hash_table_t);
    OBJ_CONSTRUCT(&mca_btl_eth_component.eth_events, opal_list_t);
    OBJ_CONSTRUCT(&mca_btl_eth_component.eth_frag_eager, ompi_free_list_t);
    OBJ_CONSTRUCT(&mca_btl_eth_component.eth_frag_max, ompi_free_list_t);
    OBJ_CONSTRUCT(&mca_btl_eth_component.eth_frag_user, ompi_free_list_t);
    opal_hash_table_init(&mca_btl_eth_component.eth_procs, 256);

    /* register ETH component parameters */
    mca_btl_eth_component.eth_num_links =
        mca_btl_eth_param_register_int("links", NULL, 1);
    mca_btl_eth_component.eth_if_include =
        mca_btl_eth_param_register_string("if_include", "Comma-delimited list of devices to use for MPI communication (e.g., \"eth0,eth1\").  Mutually exclusive with btl_eth_if_exclude.", "");
    mca_btl_eth_component.eth_if_exclude =
        mca_btl_eth_param_register_string("if_exclude", "Comma-delimited list of devices to NOT use for MPI communication -- all devices not matching these specifications will be used (e.g., \"eth0,eth1\").  Mutually exclusive with btl_eth_if_include.", "lo");
    mca_btl_eth_component.eth_free_list_num =
        mca_btl_eth_param_register_int ("free_list_num", NULL, 8);
    mca_btl_eth_component.eth_free_list_max =
        mca_btl_eth_param_register_int ("free_list_max", NULL, -1);
    mca_btl_eth_component.eth_free_list_inc =
        mca_btl_eth_param_register_int ("free_list_inc", NULL, 32);
    mca_btl_eth_component.eth_sndbuf =
        mca_btl_eth_param_register_int ("sndbuf", NULL, 128*1024);
    mca_btl_eth_component.eth_rcvbuf =
        mca_btl_eth_param_register_int ("rcvbuf", NULL, 128*1024);
    mca_btl_eth_component.eth_endpoint_cache =
        mca_btl_eth_param_register_int ("endpoint_cache",
            "The size of the internal cache for each ETH connection. This cache is"
            " used to reduce the number of syscalls, by replacing them with memcpy."
            " Every read will read the expected data plus the amount of the"
            " endpoint_cache", 30*1024);
    mca_btl_eth_component.eth_port_min =
        mca_btl_eth_param_register_int( "port_min",
            "The minimum port where the ETH BTL will try to bind (default 0)", 0 );
    mca_btl_eth_component.eth_port_range =
        mca_btl_eth_param_register_int( "port_range",
            "The number of ports where the ETH BTL will try to bind (default 64K)."
            " This parameter together with the port min, define a range of ports"
            " where Open MPI will open sockets.",
            64*1024 - mca_btl_eth_component.eth_port_min - 1);
    mca_btl_eth_module.super.btl_exclusivity =  MCA_BTL_EXCLUSIVITY_LOW + 100;
    mca_btl_eth_module.super.btl_eager_limit = 64*1024;
    mca_btl_eth_module.super.btl_rndv_eager_limit = 64*1024;
    mca_btl_eth_module.super.btl_max_send_size = 128*1024;
    mca_btl_eth_module.super.btl_rdma_pipeline_send_length = 128*1024;
    mca_btl_eth_module.super.btl_rdma_pipeline_frag_size = INT_MAX;
    mca_btl_eth_module.super.btl_min_rdma_pipeline_size = 0;
    mca_btl_eth_module.super.btl_flags = MCA_BTL_FLAGS_PUT |
                                       MCA_BTL_FLAGS_SEND_INPLACE |
                                       MCA_BTL_FLAGS_NEED_CSUM |
                                       MCA_BTL_FLAGS_NEED_ACK |
                                       MCA_BTL_FLAGS_HETEROGENEOUS_RDMA;
    mca_btl_eth_module.super.btl_bandwidth = 100;
    mca_btl_eth_module.super.btl_latency = 100;
    mca_btl_base_param_register(&mca_btl_eth_component.super.btl_version,
            &mca_btl_eth_module.super);

    mca_btl_eth_component.eth_echo_attempts =
        mca_btl_eth_param_register_int ("echo_attempts", NULL, 3);
    mca_btl_eth_component.eth_echo_timeout =
        mca_btl_eth_param_register_int ("echo_timeout", NULL, 250*1000); /* 250 ms */

    return OMPI_SUCCESS;
}


/*
 * module cleanup - sanity checking of queue lengths
 */

int mca_btl_eth_component_close(void)
{
    opal_list_item_t* item;
    opal_list_item_t* next;

    if(NULL != mca_btl_eth_component.eth_if_include) {
        free(mca_btl_eth_component.eth_if_include);
        mca_btl_eth_component.eth_if_include = NULL;
    }
    if(NULL != mca_btl_eth_component.eth_if_exclude) {
       free(mca_btl_eth_component.eth_if_exclude);
       mca_btl_eth_component.eth_if_exclude = NULL;
    }
    if (NULL != mca_btl_eth_component.eth_btls) {
        free(mca_btl_eth_component.eth_btls);
        mca_btl_eth_component.eth_btls = NULL;
    }
 
    if (mca_btl_eth_component.eth_listen_sd >= 0) {
        opal_event_del(&mca_btl_eth_component.eth_recv_event);
        CLOSE_THE_SOCKET(mca_btl_eth_component.eth_listen_sd);
        mca_btl_eth_component.eth_listen_sd = -1;
    }

    /* cleanup any pending events */
    OPAL_THREAD_LOCK(&mca_btl_eth_component.eth_lock);
    for(item =  opal_list_remove_first(&mca_btl_eth_component.eth_events);
        item != opal_list_get_end(&mca_btl_eth_component.eth_events);
        item =  next) {
        mca_btl_eth_event_t* event = (mca_btl_eth_event_t*)item;
        next = opal_list_get_next(item);
        opal_event_del(&event->event);
        OBJ_RELEASE(event);
    }
    OPAL_THREAD_UNLOCK(&mca_btl_eth_component.eth_lock);

    /* release resources */
    OBJ_DESTRUCT(&mca_btl_eth_component.eth_procs);
    OBJ_DESTRUCT(&mca_btl_eth_component.eth_events);
    OBJ_DESTRUCT(&mca_btl_eth_component.eth_frag_eager);
    OBJ_DESTRUCT(&mca_btl_eth_component.eth_frag_max);
    OBJ_DESTRUCT(&mca_btl_eth_component.eth_frag_user);
    OBJ_DESTRUCT(&mca_btl_eth_component.eth_lock);

#ifdef __WINDOWS__
    WSACleanup();
#endif

    return OMPI_SUCCESS;
}


/*
 * Determine the hardware (MAC) address of the interface specified by if_name
 */

static inline int mca_btl_eth_ifname_to_addr(const char *if_name,
        struct ether_addr *addr)
{
    struct ifreq ifdata;

    memset(&ifdata, 0, sizeof(ifdata));
    strncpy(ifdata.ifr_name, if_name, IF_NAMESIZE);
    if (ioctl(mca_btl_eth_component.eth_listen_sd, SIOCGIFHWADDR, &ifdata) < 0)
        return OMPI_ERROR;

    *addr = ETH_ARR2ADDR(ifdata.ifr_hwaddr.sa_data);

    return OMPI_SUCCESS;
}


/*
 *  Create a btl instance and add to modules list.
 */

static int mca_btl_eth_create(int if_kindex, const char* if_name)
{
    struct mca_btl_eth_module_t* btl;
    char param[256];
    int i;

    for( i = 0; i < (int)mca_btl_eth_component.eth_num_links; i++ ) {
        btl = (struct mca_btl_eth_module_t *)malloc(sizeof(mca_btl_eth_module_t));
        if(NULL == btl)
            return OMPI_ERR_OUT_OF_RESOURCE;
        memcpy(btl, &mca_btl_eth_module, sizeof(mca_btl_eth_module));
        OBJ_CONSTRUCT(&btl->eth_endpoints, opal_list_t);
        mca_btl_eth_component.eth_btls[mca_btl_eth_component.eth_num_btls++] = btl;

        /* initialize the btl */
        btl->eth_ifkindex = (uint16_t) if_kindex;
#if MCA_BTL_ETH_STATISTICS
        btl->eth_bytes_recv = 0;
        btl->eth_bytes_sent = 0;
        btl->eth_send_handler = 0;
#endif
        mca_btl_eth_ifname_to_addr(if_name, &btl->eth_ifaddr);

        /* allow user to specify interface bandwidth */
        sprintf(param, "bandwidth_%s", if_name);
        btl->super.btl_bandwidth = mca_btl_eth_param_register_int(param, NULL, btl->super.btl_bandwidth);

        /* allow user to override/specify latency ranking */
        sprintf(param, "latency_%s", if_name);
        btl->super.btl_latency = mca_btl_eth_param_register_int(param, NULL, btl->super.btl_latency);
        if( i > 0 ) {
            btl->super.btl_bandwidth >>= 1;
            btl->super.btl_latency   <<= 1;
        }

        /* allow user to specify interface bandwidth */
        sprintf(param, "bandwidth_%s:%d", if_name, i);
        btl->super.btl_bandwidth = mca_btl_eth_param_register_int(param, NULL, btl->super.btl_bandwidth);

        /* allow user to override/specify latency ranking */
        sprintf(param, "latency_%s:%d", if_name, i);
        btl->super.btl_latency = mca_btl_eth_param_register_int(param, NULL, btl->super.btl_latency);
#if 0 && OMPI_ENABLE_DEBUG
        BTL_OUTPUT(("interface %s instance %i: bandwidth %d latency %d\n", if_name, i,
                    btl->super.btl_bandwidth, btl->super.btl_latency));
#endif
    }
    return OMPI_SUCCESS;
}

/*
 * Create a ETH BTL instance for either:
 * (1) all interfaces specified by the user
 * (2) all available interfaces 
 * (3) all available interfaces except for those excluded by the user
 */

static int mca_btl_eth_component_create_instances(void)
{
    const int if_count = opal_ifcount();
    int if_index;
    int kif_count = 0;
    int *kindexes = NULL; /* this array is way too large, but never too small */
    char **include;
    char **exclude;
    char **argv;
    int ret = OMPI_SUCCESS;

    if(if_count <= 0) {
        return OMPI_ERROR;
    }

    kindexes = (int *) malloc(sizeof(int) * if_count);
    if (NULL == kindexes) {
        return OMPI_ERR_OUT_OF_RESOURCE;
    }

    /* calculate the number of kernel indexes (number of physical NICs) */
    {
        int j;

        /* initialize array to 0. Assumption: 0 isn't a valid kernel index */
        memset (kindexes, 0, sizeof(int) * if_count);

        /* assign the corresponding kernel indexes for all opal_list indexes
         * (loop over all addresses)
         */
        for(if_index = opal_ifbegin(); if_index >= 0; if_index = opal_ifnext(if_index)){
            int index = opal_ifindextokindex (if_index);
            if (index > 0) {
                bool already_seen = false;
                for (j=0; (false == already_seen) && (j < kif_count); j++) {
                    if (kindexes[j] == index) {
                        already_seen = true;
                    }
                }

                if (false == already_seen) {
                    kindexes[kif_count] = index;
                    kif_count++;
                }
            }
        }
    }

    /* allocate memory for btls */
    mca_btl_eth_component.eth_btls = (mca_btl_eth_module_t**)malloc(mca_btl_eth_component.eth_num_links *
                                                                    kif_count * sizeof(mca_btl_eth_module_t*));
    if(NULL == mca_btl_eth_component.eth_btls) {
        ret = OMPI_ERR_OUT_OF_RESOURCE;
        goto cleanup;
    }

    mca_btl_eth_component.eth_addr_count = if_count;

    /* if the user specified an interface list - use these exclusively */
    argv = include = opal_argv_split(mca_btl_eth_component.eth_if_include,',');
    while(argv && *argv) {
        char* if_name = *argv;
        int if_index = opal_ifnametokindex(if_name);
        if(if_index < 0) {
            BTL_ERROR(("invalid interface \"%s\"", if_name));
        }
        mca_btl_eth_create(if_index, if_name);
        argv++;
    }
    opal_argv_free(include);
    if(mca_btl_eth_component.eth_num_btls) {
        ret = OMPI_SUCCESS;
        goto cleanup;
    }

    /* if the interface list was not specified by the user, create 
     * a BTL for each interface that was not excluded.
    */
    exclude = opal_argv_split(mca_btl_eth_component.eth_if_exclude,',');
    {
        int i;
        for(i = 0; i < kif_count; i++) {
            /* Bug, FIXME: Don't hardcode length of if_name, use IFNAMESIZE */
            char if_name[32];
            if_index = kindexes[i];

            opal_ifkindextoname(if_index, if_name, sizeof(if_name));

            /* check to see if this interface exists in the exclude list */
            argv = exclude;
            while(argv && *argv) {
                if(strncmp(*argv,if_name,strlen(*argv)) == 0)
                    break;
                argv++;
            }
            /* if this interface was not found in the excluded list, create a BTL */
            if(argv == 0 || *argv == 0) {
                mca_btl_eth_create(if_index, if_name);
            }
        }
    }
    opal_argv_free(exclude);

 cleanup:
    if (NULL != kindexes) {
        free(kindexes);
    }
    return ret;
}

/*
 * Create a listen socket and bind to all interfaces
 */

static int mca_btl_eth_component_create_listen(void)
{
    int flags;
    int sd;
    struct sockaddr_en enaddr;
    opal_socklen_t addrlen;

    /* create a listen socket for incoming connections */
    sd = socket(AF_ENET, SOCK_STREAM, 0);
    if(sd < 0) {
        if (EAFNOSUPPORT != opal_socket_errno) {
            BTL_ERROR(("socket() failed: %s (%d)",
                       strerror(opal_socket_errno), opal_socket_errno));
        }
        return OMPI_ERR_IN_ERRNO;
    }

    mca_btl_eth_set_socket_options(sd);

    memset(&enaddr, 0, sizeof(enaddr));
    ((struct sockaddr_en*) &enaddr)->sen_family = AF_ENET;
    addrlen = sizeof(struct sockaddr_en);

    {  /* Don't reuse ports */
        int flg = 0;
        if (setsockopt (sd, SOL_SOCKET, SO_REUSEADDR, (const char *)&flg, sizeof (flg)) < 0) {
            BTL_ERROR(("mca_btl_eth_create_listen: unable to unset the "
                           "SO_REUSEADDR option (%s:%d)\n",
                           strerror(opal_socket_errno), opal_socket_errno));
            CLOSE_THE_SOCKET(sd);
            return OMPI_ERROR;
        }
    }

    {
        int index, range, port;
        
        range = mca_btl_eth_component.eth_port_range;
        port = mca_btl_eth_component.eth_port_min;

        for( index = 0;  index < range; index++ ) {
            ((struct sockaddr_en*) &enaddr)->sen_port = htons(port + index);
            if(bind(sd, (struct sockaddr*)&enaddr, addrlen) < 0) {
                if( (EADDRINUSE == opal_socket_errno) || (EADDRNOTAVAIL == opal_socket_errno) ) {
                    continue;
                }
                BTL_ERROR(("bind() failed: %s (%d)",
                          strerror(opal_socket_errno), opal_socket_errno));
                CLOSE_THE_SOCKET(sd);
                return OMPI_ERROR;
            }
            goto socket_binded;
        }
        BTL_ERROR(("bind() failed: no port available in the range [%d..%d]",
                   mca_btl_eth_component.eth_port_min,
                   mca_btl_eth_component.eth_port_min + range));
        CLOSE_THE_SOCKET(sd);
        return OMPI_ERROR;
    }
  socket_binded:
    /* resolve system assignend port */
    if(getsockname(sd, (struct sockaddr*)&enaddr, &addrlen) < 0) {
        BTL_ERROR(("getsockname() failed: %s (%d)",
                   strerror(opal_socket_errno), opal_socket_errno));
        CLOSE_THE_SOCKET(sd);
        return OMPI_ERROR;
    }

    mca_btl_eth_component.eth_listen_port = ((struct sockaddr_en*) &enaddr)->sen_port;
    mca_btl_eth_component.eth_listen_sd = sd;

    /* setup listen backlog to maximum allowed by kernel */
    if(listen(sd, SOMAXCONN) < 0) {
        BTL_ERROR(("listen() failed: %s (%d)", 
                   strerror(opal_socket_errno), opal_socket_errno));
        CLOSE_THE_SOCKET(sd);
        return OMPI_ERROR;
    }

    /* set socket up to be non-blocking, otherwise accept could block */
    if((flags = fcntl(sd, F_GETFL, 0)) < 0) {
        BTL_ERROR(("fcntl(F_GETFL) failed: %s (%d)",
                   strerror(opal_socket_errno), opal_socket_errno));
        CLOSE_THE_SOCKET(sd);
        return OMPI_ERROR;
    } else {
        flags |= O_NONBLOCK;
        if(fcntl(sd, F_SETFL, flags) < 0) {
            BTL_ERROR(("fcntl(F_SETFL) failed: %s (%d)",
                       strerror(opal_socket_errno), opal_socket_errno));
            CLOSE_THE_SOCKET(sd);
            return OMPI_ERROR;
        }
    }

    /* register listen port */
    opal_event_set( &mca_btl_eth_component.eth_recv_event,
                    mca_btl_eth_component.eth_listen_sd,
                    OPAL_EV_READ|OPAL_EV_PERSIST,
                    mca_btl_eth_component_accept_handler,
                    0 );
    opal_event_add(&mca_btl_eth_component.eth_recv_event, 0);
    return OMPI_SUCCESS;
}

/*
 *  Register ETH module addressing information. The MCA framework
 *  will make this available to all peers. 
 */

static int mca_btl_eth_component_exchange(void)
{
     int rc = 0;
     size_t i = 0;
     size_t size = mca_btl_eth_component.eth_addr_count * 
                   mca_btl_eth_component.eth_num_links * sizeof(mca_btl_eth_addr_t);
     size_t xfer_size = 0; /* real size to transfer (may differ from 'size') */

     if(mca_btl_eth_component.eth_num_btls != 0) {
         mca_btl_eth_addr_t *addrs = (mca_btl_eth_addr_t *)malloc(size);
         memset(addrs, 0, size);

         /* here we start populating our addresses */
         for(i=0; i<mca_btl_eth_component.eth_num_btls; i++) {
             struct mca_btl_eth_module_t* btl = mca_btl_eth_component.eth_btls[i];
             addrs[i].addr_enet  = btl->eth_ifaddr;
             addrs[i].addr_port  = mca_btl_eth_component.eth_listen_port;
             addrs[i].addr_inuse = 0;
             xfer_size += sizeof (mca_btl_eth_addr_t);
         }
         rc =  ompi_modex_send(&mca_btl_eth_component.super.btl_version, 
                               addrs, xfer_size);
         free(addrs);
     } /* end if */
     return rc;
}

/*
 *  ETH module initialization:
 *  (1) read interface list from kernel and compare against module parameters
 *      then create a BTL instance for selected interfaces
 *  (2) setup ETH listen socket for incoming connection attempts
 *  (3) register BTL parameters with the MCA
 */
mca_btl_base_module_t** mca_btl_eth_component_init(int *num_btl_modules, 
                                                   bool enable_progress_threads,
                                                   bool enable_mpi_threads)
{
    int ret = OMPI_SUCCESS;
    mca_btl_base_module_t **btls;
    *num_btl_modules = 0;

    /* initialize free lists */
    ompi_free_list_init_new( &mca_btl_eth_component.eth_frag_eager,
                         sizeof (mca_btl_eth_frag_eager_t) + 
                         mca_btl_eth_module.super.btl_eager_limit,
                         CACHE_LINE_SIZE,
                         OBJ_CLASS (mca_btl_eth_frag_eager_t),
                         0,CACHE_LINE_SIZE,
                         mca_btl_eth_component.eth_free_list_num,
                         mca_btl_eth_component.eth_free_list_max,
                         mca_btl_eth_component.eth_free_list_inc,
                         NULL );
                                                                                                                                  
    ompi_free_list_init_new( &mca_btl_eth_component.eth_frag_max,
                         sizeof (mca_btl_eth_frag_max_t) + 
                         mca_btl_eth_module.super.btl_max_send_size,
                         CACHE_LINE_SIZE,
                         OBJ_CLASS (mca_btl_eth_frag_max_t),
                         0,CACHE_LINE_SIZE,
                         mca_btl_eth_component.eth_free_list_num,
                         mca_btl_eth_component.eth_free_list_max,
                         mca_btl_eth_component.eth_free_list_inc,
                         NULL );
                                                                                                                                  
    ompi_free_list_init_new( &mca_btl_eth_component.eth_frag_user,
                         sizeof (mca_btl_eth_frag_user_t),
                         CACHE_LINE_SIZE,
                         OBJ_CLASS (mca_btl_eth_frag_user_t),
                         0,CACHE_LINE_SIZE,
                         mca_btl_eth_component.eth_free_list_num,
                         mca_btl_eth_component.eth_free_list_max,
                         mca_btl_eth_component.eth_free_list_inc,
                         NULL );
                                                                                                                                  
    /* create a ETH listen socket for incoming connection attempts 
     * this is now done before module creation so that the modules
     * can use this socket for hardware address determination
     * (it opens the theoretical possibility of connection attempts
     * before any module is created, but practically there should be
     * none until we have published our address) */
    if(OMPI_SUCCESS != (ret = mca_btl_eth_component_create_listen() )) {
        return 0;
    }

    /* create a BTL ETH module for selected interfaces */
    if(OMPI_SUCCESS != (ret = mca_btl_eth_component_create_instances() )) {
        return 0;
    }

    /* publish ETH parameters with the MCA framework */
    if(OMPI_SUCCESS != (ret = mca_btl_eth_component_exchange() )) {
        return 0;
    }

    btls = (mca_btl_base_module_t **)malloc(mca_btl_eth_component.eth_num_btls * 
                  sizeof(mca_btl_base_module_t*));
    if(NULL == btls) {
        return NULL;
    }

    memcpy(btls, mca_btl_eth_component.eth_btls, mca_btl_eth_component.eth_num_btls*sizeof(mca_btl_eth_module_t*));
    *num_btl_modules = mca_btl_eth_component.eth_num_btls;
    return btls;
}

/*
 *  ETH module control
 */

int mca_btl_eth_component_control(int param, void* value, size_t size)
{
    return OMPI_SUCCESS;
}


/**
 * Called by the event engine when the listening socket has
 * a connection event. Accept the incoming connection request
 * and queue them for completion of the connection handshake.
 */
static void mca_btl_eth_component_accept_handler( int incoming_sd,
                                                  short ignored,
                                                  void* unused )
{
    while(true) {
        struct sockaddr_en addr;
        opal_socklen_t addrlen = sizeof(addr);

        mca_btl_eth_event_t *event;
        int sd = accept(incoming_sd, (struct sockaddr*)&addr, &addrlen);
        if(sd < 0) {
            if(opal_socket_errno == EINTR)
                continue;
            if(opal_socket_errno != EAGAIN && opal_socket_errno != EWOULDBLOCK)
                BTL_ERROR(("accept() failed: %s (%d).", 
                           strerror(opal_socket_errno), opal_socket_errno));
            return;
        }
        mca_btl_eth_set_socket_options(sd);

        /* wait for receipt of peers process identifier to complete this connection */
         
        event = OBJ_NEW(mca_btl_eth_event_t);
        opal_event_set(&event->event, sd, OPAL_EV_READ, mca_btl_eth_component_recv_handler, event);
        opal_event_add(&event->event, 0);
    }
}


/**
 * Event callback when there is data available on the registered 
 * socket to recv. This callback is triggered only once per lifetime
 * for any socket, in the beginning when we setup the handshake
 * protocol.
 */
static void mca_btl_eth_component_recv_handler(int sd, short flags, void* user)
{
    orte_process_name_t guid;
    struct sockaddr_storage addr;
    int retval;
    mca_btl_eth_proc_t* btl_proc;
    opal_socklen_t addr_len = sizeof(addr);
    mca_btl_eth_event_t *event = (mca_btl_eth_event_t *)user;

    OBJ_RELEASE(event);

    /* recv the process identifier */
    retval = recv(sd, (char *)&guid, sizeof(guid), 0);
    if(retval != sizeof(guid)) {
        CLOSE_THE_SOCKET(sd);
        return;
    }
    ORTE_PROCESS_NAME_NTOH(guid);

    /* now set socket up to be non-blocking */
    if((flags = fcntl(sd, F_GETFL, 0)) < 0) {
        BTL_ERROR(("fcntl(F_GETFL) failed: %s (%d)",
                   strerror(opal_socket_errno), opal_socket_errno));
    } else {
        flags |= O_NONBLOCK;
        if(fcntl(sd, F_SETFL, flags) < 0) {
            BTL_ERROR(("fcntl(F_SETFL) failed: %s (%d)",
                       strerror(opal_socket_errno), opal_socket_errno));
        }
    }
   
    /* lookup the corresponding process */
    btl_proc = mca_btl_eth_proc_lookup(&guid);
    if(NULL == btl_proc) {
        CLOSE_THE_SOCKET(sd);
        return;
    }

    /* lookup peer address */
    if(getpeername(sd, (struct sockaddr*)&addr, &addr_len) != 0) {
        BTL_ERROR(("getpeername() failed: %s (%d)", 
                   strerror(opal_socket_errno), opal_socket_errno));
        CLOSE_THE_SOCKET(sd);
        return;
    }

    /* are there any existing peer instances will to accept this connection */
    if(mca_btl_eth_proc_accept(btl_proc, (struct sockaddr_en*)&addr, sd) == false) {
        CLOSE_THE_SOCKET(sd);
        return;
    }
}

