/*
 * Copyright (c) 2006 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 * Copyright (c) 2006 The Technical University of Chemnitz. All 
 *                    rights reserved.
 *
 * Author(s): (alphabetically)
 *            Sven Eckelmann
 *            Torsten Hoefler 
 *            Mirko Reinhard 
 *            Matthias Treydte 
 */

#ifndef ENET_DGRAM_H_
#define ENET_DGRAM_H_

#include <linux/version.h>
#include <linux/list.h>  /* list structs and ops */
#include <linux/proc_fs.h> /* struct file_operations */

#include "enet_compat.h"


/*
 * Ethernet packet type to use for EDP frames
 *
 * (choose a packet type currently unused by the kernel, see
 *  ETH_P_* constants in %KERNEL_SRC%/include/linux/if_ether.h
 *  or http://www.iana.org/assignments/ethernet-numbers)
 *
 * Furthermore, the packet type should be larger than or equal to
 * 0x0600 or 1536. Otherwise the linux kernel assumes that this
 * value denotes the frame length of an IEEE-802.2 / 802.3 ethernet
 * frame, which have a slightly different structure compared to the
 * 'normal' blue-book ethernet frame.
 */
#define ETH_P_EDP 0x8889


/* EDP packet header */
struct edp_header {
	unsigned short dst_port; /**< edp destination port */
	unsigned short src_port; /**< edp source port */
	unsigned short length;  /**< data length */
} __attribute__((packed));


/* EDP header length */
#define EDP_HLEN sizeof(struct edp_header)


/* Private EDP socket data structure */
struct edp_opt {
	struct edp_stats stats;
	int   ifindex;  /**< bound interface */
	short int  port;   /**< bound port (host byte order) */
	short int  sport;   /**< bound source port (net byte order) */
	unsigned char  connected;  /**< whether the socket is connected */
	short int  dport;   /**< connect()'ed destintion port (net byte order) */
	unsigned char  daddr[ETH_ALEN]; /**< connect()'ed destination address */
};


/* EDP socket data structure */
struct edp_sock {
	struct sock  sk; /**< basic kernel socket struct */
	struct edp_opt  edp; /**< edp socket options */
};


/* access private EDP socket information */
#define EDP_SK(__sk) \
 (&((struct edp_sock *)(__sk))->edp)


/* counter of EDP sockets alive */
extern atomic_t edp_skcount;


/* List of all EDP sockets and its lock */
extern struct hlist_head edp_sklist;
extern rwlock_t edp_sklist_lock;


/* EDP socket buffer control block */
struct edp_skb_cb {
	unsigned int ifindex; /*< interface index */
	unsigned char addrlen; /*< address length */
	unsigned char addr[8]; /*< link layer address */
};


/* access the private socket buffer control block data */
#define EDP_SKB_CB(__skb) ((struct edp_skb_cb *)&((__skb)->cb[0]))



/*
 * EDP ports
 */

/* port-to-socket hashtable size
 * (define a hashtable size of 2^x - it is often used like:  port & (EDP_HTABLE_SIZE - 1) ) */
#define EDP_HTABLE_SIZE 128


/* port-to-socket hashtable */
extern struct hlist_head edp_hash[EDP_HTABLE_SIZE];


/* port-to-socket hashtable r/w-lock */
extern rwlock_t edp_hash_lock;


/* port range to automatically choose ports from and port rover */
extern int edp_local_port_range[2];
extern int edp_port_rover;


/* determines whether a given port is already in use */
static inline int edp_port_inuse(u16 port, int ifindex)
{
	struct sock *sk;
	struct hlist_node *node;

	sk_for_each_bound(sk, node, &edp_hash[port & (EDP_HTABLE_SIZE - 1)]) {
		if ((EDP_SK(sk)->port == port) && (!EDP_SK(sk)->ifindex ||
		                                   !ifindex || EDP_SK(sk)->ifindex == ifindex))
			return 1;
	}
	return 0;
}


/**
 * EDP_HTABLE_SK_GET - get an EDP socket for an interface/port
 * @sk:  &struct sock to find
 * @pos: &struct hlist_node loop counter
 * @htable: hashtable
 * @hsize: int hash table size
 * @iface: int interface index
 * @hport: int port (host byte order)
 *
 * The hashtable r/w lock must be held for reading
 */
#define EDP_HTABLE_SK_GET(sk, pos, htable, hsize, iface, hport)    \
 hlist_for_each(pos, &htable[hport & (hsize - 1)]) {    \
  struct sock *__sk = hlist_entry(pos, struct sock, sk_bind_node); \
  struct edp_opt *__esk = EDP_SK(__sk);     \
  if (__esk->port == hport &&      \
      (__esk->ifindex == 0 || __esk->ifindex == iface)) {   \
   sk = __sk;       \
   sock_hold(sk);       \
   break;        \
  }         \
 }


/*
 * EDP socket operation prototypes
 */
int edp_prot_init(void);
void edp_prot_destroy(void);
struct sock *sock_edp_create(NET_PARAM struct socket *sock, int protocol, int *err);


/*
 * EDP kernel hook data structures
 */
extern struct packet_type     edp_receive_hook;
extern struct notifier_block  edp_netdev_notifier;

#ifdef CONFIG_PROC_FS
extern struct file_operations edp_proc_fops;
#endif


#endif /*ENET_DGRAM_H_*/
