/*
 * Copyright (c) 2006 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 * Copyright (c) 2006 The Technical University of Chemnitz. All 
 *                    rights reserved.
 *
 * Author(s): (alphabetically)
 *            Sven Eckelmann
 *            Torsten Hoefler 
 *            Mirko Reinhard 
 *            Matthias Treydte 
 */

#include "enet_stream_cong.h"
#include <linux/jiffies.h>
#include <linux/ctype.h>
#include "enet_compat.h"

/* declare some variables */
int esp_burst_length = ESP_INITIAL_BURST_LENGTH;
int esp_initial_ack_burst_length = ESP_INITIAL_ACK_INITIAL_BURST_LENGTH;
int esp_packets_to_ack = ESP_INITIAL_PACKETS_TO_ACK;
int esp_round_trip_time = ESP_INITIAL_ROUND_TRIP_TIME;
int esp_default_rcvbuf = ESP_INITIAL_RCVBUF;
int esp_default_sndbuf = ESP_INITIAL_SNDBUF;
struct sk_buff_head esp_ack_queue;
atomic_t esp_recv_count;

void esp_enter_rcv_mem_pressure(struct sock *sk)
{
	struct esp_opt *esp_sk = ESP_SK(sk);

	/* take care not to do this twice */
	if (esp_sk->rcv_mem_pressure)
		return;

	PR_DEBUG("Socket %p entering recv memory pressure mode.", sk);

	esp_sk->rcv_mem_pressure = 1;
	atomic_dec(&esp_recv_count);
	esp_rrq_clear(sk);
	esp_purge_ack(sk);
}

void esp_lift_rcv_mem_pressure(struct sock *sk)
{
	struct esp_opt *esp_sk = ESP_SK(sk);

	/* take care not to do this twice */
	if (!esp_sk->rcv_mem_pressure)
		return;

	PR_DEBUG("Socket %p leaves recv memory pressure mode.", sk);

	atomic_inc(&esp_recv_count);
	esp_sk->rcv_mem_pressure = 0;
}

/* older kernels don't have these two functions. don't know
 * when exactly they were introduced, though. 2.6.16 got them.
 * and in kernel version 2.6.9 they are defined in another
 * header file. i'm fed up!
 */

static inline unsigned int my_jiffies_to_usecs(const unsigned long j)
{
#if HZ <= USEC_PER_SEC && !(USEC_PER_SEC % HZ)
	return (USEC_PER_SEC / HZ) * j;
#elif HZ > USEC_PER_SEC && !(HZ % USEC_PER_SEC)
	return (j + (HZ / USEC_PER_SEC) - 1) / (HZ / USEC_PER_SEC);
#else
	return (j * USEC_PER_SEC) / HZ;
#endif
}

static inline unsigned long my_usecs_to_jiffies(const unsigned int u)
{
	if (u > jiffies_to_usecs(MAX_JIFFY_OFFSET))
		return MAX_JIFFY_OFFSET;
#if HZ <= USEC_PER_SEC && !(USEC_PER_SEC % HZ)
	return (u + (USEC_PER_SEC / HZ) - 1) / (USEC_PER_SEC / HZ);
#elif HZ > USEC_PER_SEC && !(HZ % USEC_PER_SEC)
	return u * (HZ / USEC_PER_SEC);
#else
	return (u * HZ + USEC_PER_SEC - 1) / USEC_PER_SEC;
#endif
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
static int rtt_handler(struct ctl_table *table, int __user *name, int nlen,
                       void __user *oldval, size_t __user *oldlenp,
                       void __user *newval, size_t newlen)
#else
static int rtt_handler(struct ctl_table *table, int __user *name, int nlen,
                       void __user *oldval, size_t __user *oldlenp,
                       void __user *newval, size_t newlen,
                       void **context)
#endif
{
	size_t size;
	unsigned int rtt;

	if (oldval) {
		/* we should return the old value */
		get_user(size, oldlenp);
		if (size < sizeof(unsigned int))
			return -EINVAL;

		put_user(sizeof(unsigned int), oldlenp);
		rtt = jiffies_to_usecs(esp_round_trip_time);
		put_user(rtt, (unsigned int __user *)oldval);
	}

	if (newval) {
		/* we should set the new value */
		if (newlen < sizeof(unsigned int))
			return -EINVAL;

		get_user(rtt, (unsigned int __user *)newval);
		esp_round_trip_time = my_usecs_to_jiffies(rtt);
	}

	return 1;
}



/**
 * just copied over from sysctl.c as it's not exported by the kernel.
 * BEWARE: the conv - parameter MUST be non-NULL in this version.
 */
static int do_proc_dointvec(ctl_table *table, int write, struct file *filp,
                            void __user *buffer, size_t *lenp, loff_t *ppos,
                            int (*conv)(int *negp, unsigned long *lvalp, int *valp,
                                        int write, void *data),
                            void *data)
{
#define TMPBUFLEN 21
	int *i, vleft, first = 1, neg, val;
	unsigned long lval;
	size_t left, len;

	char buf[TMPBUFLEN], *p;
	char __user *s = buffer;

	if (!table->data || !table->maxlen || !*lenp ||
	                (*ppos && !write)) {
		*lenp = 0;
		return 0;
	}

	i = (int *) table->data;
	vleft = table->maxlen / sizeof(*i);
	left = *lenp;

	for (; left && vleft--; i++, first = 0) {
		if (write) {
			while (left) {
				char c;
				if (get_user(c, s))
					return -EFAULT;
				if (!isspace(c))
					break;
				left--;
				s++;
			}
			if (!left)
				break;
			neg = 0;
			len = left;
			if (len > sizeof(buf) - 1)
				len = sizeof(buf) - 1;
			if (copy_from_user(buf, s, len))
				return -EFAULT;
			buf[len] = 0;
			p = buf;
			if (*p == '-' && left > 1) {
				neg = 1;
				left--, p++;
			}
			if (*p < '0' || *p > '9')
				break;

			lval = simple_strtoul(p, &p, 0);

			len = p - buf;
			if ((len < left) && *p && !isspace(*p))
				break;
			if (neg)
				val = -val;
			s += len;
			left -= len;

			if (conv(&neg, &lval, i, 1, data))
				break;
		} else {
			p = buf;
			if (!first)
				*p++ = '\t';

			if (conv(&neg, &lval, i, 0, data))
				break;

			sprintf(p, "%s%lu", neg ? "-" : "", lval);
			len = strlen(buf);
			if (len > left)
				len = left;
			if (copy_to_user(s, buf, len))
				return -EFAULT;
			left -= len;
			s += len;
		}
	}

	if (!write && !first && left) {
		if (put_user('\n', s))
			return -EFAULT;
		left--, s++;
	}
	if (write) {
		while (left) {
			char c;
			if (get_user(c, s++))
				return -EFAULT;
			if (!isspace(c))
				break;
			left--;
		}
	}
	if (write && first)
		return -EINVAL;
	*lenp -= left;
	*ppos += *lenp;
	return 0;
#undef TMPBUFLEN
}

static int do_proc_dointvec_us_jiffies_conv(int *negp, unsigned long *lvalp,
                int *valp,
                int write, void *data)
{
	if (write) {
		*valp = my_usecs_to_jiffies(*negp ? -*lvalp : *lvalp);
	} else {
		int val = *valp;
		unsigned long lval;
		if (val < 0) {
			*negp = -1;
			lval = (unsigned long) - val;
		} else {
			*negp = 0;
			lval = (unsigned long)val;
		}
		*lvalp = jiffies_to_usecs(lval);
	}
	return 0;
}

static int proc_dointvec_us_jiffies(ctl_table *table, int write, struct file *filp,
                                    void __user *buffer, size_t *lenp, loff_t *ppos)
{
	return do_proc_dointvec(table, write, filp, buffer, lenp, ppos,
	                        do_proc_dointvec_us_jiffies_conv, NULL);
}

#define RIGHTS 0666

static struct ctl_table esp_options_table[] = {
	{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
		.ctl_name     = CTL_UNNUMBERED,
#else
		.ctl_name     = CTL_BURST_LENGTH,
#endif
		.procname     = "burst_length",
		.data         = &esp_burst_length,
		.maxlen       = sizeof(esp_burst_length),
		.mode         = RIGHTS,
		.proc_handler = &proc_dointvec
	},
	{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
		.ctl_name     = CTL_UNNUMBERED,
#else
		.ctl_name     = CTL_INITIAL_ACK_BURST_LENGTH,
#endif
		.procname     = "initial_ack_burst_length",
		.data         = &esp_initial_ack_burst_length,
		.maxlen       = sizeof(esp_initial_ack_burst_length),
		.mode         = RIGHTS,
		.proc_handler = &proc_dointvec
	},
	{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
		.ctl_name     = CTL_UNNUMBERED,
#else
		.ctl_name     = CTL_PACKETS_TO_ACK,
#endif
		.procname     = "packets_to_ack",
		.data         = &esp_packets_to_ack,
		.maxlen       = sizeof(esp_packets_to_ack),
		.mode         = RIGHTS,
		.proc_handler = &proc_dointvec
	},
	{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
		.ctl_name     = CTL_UNNUMBERED,
#else
		.ctl_name     = CTL_ROUND_TRIP_TIME,
#endif
		.procname     = "round_trip_time",
		.data         = &esp_round_trip_time,
		.maxlen       = sizeof(esp_round_trip_time),
		.mode         = RIGHTS,
		.proc_handler = &proc_dointvec_us_jiffies,
		.strategy     = &rtt_handler
	},
	{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
		.ctl_name     = CTL_UNNUMBERED,
#else
		.ctl_name     = CTL_SEND_BUFF,
#endif
		.procname     = "send_buff_size",
		.data         = &esp_default_sndbuf,
		.maxlen       = sizeof(esp_default_sndbuf),
		.mode         = RIGHTS,
		.proc_handler = &proc_dointvec
	},
	{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
		.ctl_name     = CTL_UNNUMBERED,
#else
		.ctl_name     = CTL_RECV_BUFF,
#endif
		.procname     = "recv_buff_size",
		.data         = &esp_default_rcvbuf,
		.maxlen       = sizeof(esp_default_rcvbuf),
		.mode         = RIGHTS,
		.proc_handler = &proc_dointvec
	},
	{ 0 }
};

struct ctl_table esp_ctl_table[] = {
	{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
		.ctl_name     = CTL_UNNUMBERED,
#else
		.ctl_name = CTL_ESP,
#endif
		.procname = "esp",
		.mode     = 0555,
		.child    = esp_options_table
	},
	{ 0 }
};

