/*
 * Copyright (c) 2006 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 * Copyright (c) 2006 The Technical University of Chemnitz. All 
 *                    rights reserved.
 *
 * Author(s): (alphabetically)
 *            Sven Eckelmann
 *            Torsten Hoefler 
 *            Mirko Reinhard 
 *            Matthias Treydte 
 */

#include "enet_stream_input.h"
#include "enet_module.h"
#include "enet_compat.h"

/*
 * Central data receiving function,
 * called from the kernel hook function esp_rcv_hook() and while backlog
 * processing via sk->sk_backlog_rcv().
 * The socket is locked when this function is called and not in the user's
 * hands, so socket state changes are possible.
 */
int esp_rcv_packet(struct sock *sk, struct sk_buff *skb)
{
	struct esp_opt *esp_sk = ESP_SK(sk);
	struct esp_header *esp_hdr = ESP_HDR(skb);
	struct sock *child;
	int err = 0;

	PR_DEBUG("esp socket %p state %s processing received packet "
	         "seq %d ack %d "ESP_FLAGS_FORMAT,
	         sk, esp_statenames[sk->sk_state], ESP_SKB_CB(skb)->pkt_seq,
	         ESP_SKB_CB(skb)->ack_seq, ESP_FLAGS_DEBUG(esp_hdr->flags));

	switch (sk->sk_state) {
	case ESP_ESTABLISHED:
		/* discard old packets and notify the sender */
		if (seq_before(ESP_SKB_CB(skb)->pkt_seq, esp_sk->recv_seq)) {
			//esp_send_ack(sk);
			goto out_discard;
		}

		/* check for RST */
		if (esp_hdr->flags & ESP_FLAG_RST) {
			esp_process_rst(sk, skb, -ECONNRESET);
			goto out_discard;
		}

		/* check for SYN */
		if (esp_hdr->flags & ESP_FLAG_SYN) {
			PR_DEBUG("esp socket %p received unexpected SYN - "
			         "resetting connection", sk);
			SK_ERR_REPORT(sk, -ECONNRESET);
			esp_set_state(sk, ESP_CLOSE);
			__skb_queue_purge(&sk->sk_write_queue);
			goto out_reset;
		}

		/* process received ACK */
		if (esp_hdr->flags & ESP_FLAG_ACK) {
			if (esp_process_ack_early(sk, skb)) {
				if (!(esp_hdr->flags & ESP_FLAG_RRQ)) {
					esp_push_pending_frames(sk);
				}
				esp_process_ack_late(sk, skb);
			}

			if (esp_hdr->flags & ESP_FLAG_RRQ) {
				esp_process_rrq(sk, skb);
			}

			goto out_discard;
		}

		/* check for out-of-sequence data */
		if (seq_after(ESP_SKB_CB(skb)->pkt_seq, esp_sk->recv_seq)) {
#if ESP_USE_OOS_QUEUE
			if (esp_enqueue_oos(sk, skb))
				goto out_queued;
#endif
			PR_DEBUG("At least one packet was lost! (pkt_seq %d, recv_seq %d)",
			         ESP_SKB_CB(skb)->pkt_seq, esp_sk->recv_seq);

			++esp_sk->stats.seq_lost;

			if (!esp_sk->rcv_mem_pressure) {
				/* we immediately request a re-transmission */
				printk("we request a RTX (pkt_seq: %d, recv_seq: %d)\n",
				       ESP_SKB_CB(skb)->pkt_seq, esp_sk->recv_seq);
				esp_rrq_clear(sk);
				esp_purge_ack(sk);
				esp_enqueue_ack(sk, ESP_FLAG_RRQ);
			}

			goto out_discard;
		}

		/* check for FIN */
		if (esp_hdr->flags & ESP_FLAG_FIN) {
			esp_process_fin(sk, skb);
			goto out_discard;
		}

		/* enqueue in-sequence data */
		switch (esp_enqueue_data(sk, skb)) {
		case 0: /* the packet wasn't enqueued and should be discarded */
			goto out_discard;
		case 1: /* ACK the received and enqueued data packet */
			/* maybe we should open sender's cwnd */
			if (esp_sk->recv_announced) {
				/* we received data so we can defer RRQs */
				ESP_RESET_RRQ_TIMER(sk, esp_round_trip_time);

				if (esp_should_send_ack(esp_sk)) {
					if (!esp_socket_low_rcv_mem(sk))
						esp_enqueue_ack(sk, 0);
					else
						esp_enter_rcv_mem_pressure(sk);
				}
			}

			/* if recv_announced was not set this was a
			 * one-shot sending and the process_txs_txf
			 * function will send out the ACK for us. No
			 * need do to it here.
			 */

			break;
		case -1:/* the current packet filled a hole and oos packets
       * have been processed - send ACK immediately */
			esp_send_ack(sk);
			break;
		}

		if (esp_process_txs_txf(sk, skb))
			esp_enqueue_ack(sk, 0);

		goto out_queued;

	case ESP_SYN_SENT:
		/* check for ACK */
		if (!(esp_hdr->flags & ESP_FLAG_ACK))
			goto out_discard; /* x-connect not supported */
		/* now check the ACK sequence */
		if (ESP_SKB_CB(skb)->ack_seq != esp_sk->send_seq)
			goto out_reset;
		/* check for RST */
		if (esp_hdr->flags & ESP_FLAG_RST) {
			esp_process_rst(sk, skb, -ECONNREFUSED);
			goto out_discard;
		}
		/* check for SYN (-ACK) */
		if (!(esp_hdr->flags & ESP_FLAG_SYN))
			goto out_discard;

		/* We have SYN and ACK, the packet is acceptable,
		 * now switch to ESTABLISHED state */
		WARN_ON(!esp_process_ack(sk, skb)); /* the ACK should be acceptable now */
		esp_sk->recv_seq = ESP_SKB_CB(skb)->pkt_seq + 1;
		esp_sk->ack_seq  = ESP_SKB_CB(skb)->pkt_seq;
		esp_sk->copy_seq = esp_sk->recv_seq;
		esp_send_ack(sk);
		/* immediately reply ACK so no data can be sent during
		 * connection handshake */
		esp_set_state(sk, ESP_ESTABLISHED);
		sk->sk_state_change(sk);

		/* ignore anything else */
		goto out_discard;

	case ESP_SYN_RECV:
		/* check for RST */
		if (esp_hdr->flags & ESP_FLAG_RST) {
			if (esp_process_rst(sk, skb, -ECONNRESET)) {
				sock_orphan(sk); /* sets the SOCK_DEAD socket flag */
				esp_sk_destroy(sk);
			}
			goto out_discard;
		}
		/* we're just waiting for the final ACK of our SYN-ACK to change
		 * the socket's state to ESTABLISHED */
		if (esp_hdr->flags & ESP_FLAG_ACK) {
			if (esp_process_ack(sk, skb)) {
				struct sock *parent = esp_sk->parent;

				/* put the child socket into the acceptable queue */
				bh_lock_sock(parent);
				esp_openreq_dequeue(sk);
				esp_set_state(sk, ESP_ESTABLISHED);
				esp_openreq_enqueue(parent, sk);
				/* notify all waiters on the listening parent
				* socket that there is a new accept request
				* pending */
				sk->sk_state_change(parent);
				bh_unlock_sock(parent);
			}
			goto out_discard;
		}

		/* check for SYN - retransmission */
		if (esp_hdr->flags & ESP_FLAG_SYN) {
			if (!seq_before(ESP_SKB_CB(skb)->pkt_seq, esp_sk->una_seq)) {
				PR_DEBUG("esp socket %p received unexpected SYN - "
				         "resetting connection", sk);
				SK_ERR_REPORT(sk, -ECONNRESET);
				esp_set_state(sk, ESP_CLOSE);
				__skb_queue_purge(&sk->sk_write_queue);
				goto out_reset;
			}
			/* if our SYN-ACK got lost, it will be retransmitted */
			goto out_discard;
		}
		/* nothing else allowed for now... */
		goto out_reset;

	case ESP_LISTEN:
		/* check for ACK */
		if (esp_hdr->flags & ESP_FLAG_ACK)
			goto out_reset;
		/* check for RST */
		if (esp_hdr->flags & ESP_FLAG_RST)
			goto out_discard;
		/* check for SYN */
		if (esp_hdr->flags & ESP_FLAG_SYN) {
			/* check syn and accept queues */
			if (esp_synq_is_full(sk))
				goto out_discard;
			child = esp_sk_spawn_child(sk, skb);
			if (!child)
				goto out_discard;
			/* send a SYN-ACK via the new child socket */
			esp_send_syn(child, SKB_INPUT_DEV(skb), GFP_ATOMIC);
		}
		/* ignore anything else */
		goto out_discard;

	case ESP_FIN_WAIT1:
	case ESP_FIN_WAIT2:
		/* discard old packets and notify the sender */
		if (seq_before(ESP_SKB_CB(skb)->pkt_seq, esp_sk->recv_seq)) {
			esp_send_ack(sk);
			goto out_discard;
		}
		/* check for RST */
		if (esp_hdr->flags & ESP_FLAG_RST) {
			esp_process_rst(sk, skb, -ECONNRESET);
			goto out_discard;
		}
		/* process received ACK */
		if ((sk->sk_state == ESP_FIN_WAIT1) &&
		                (esp_hdr->flags & ESP_FLAG_ACK)) {
			esp_process_ack(sk, skb);
			/* change state when everything has been ACK'ed */
			if (esp_sk->una_seq == esp_sk->send_seq) {
				esp_set_state(sk, ESP_FIN_WAIT2);
				sk->sk_shutdown |= SEND_SHUTDOWN;
				if (!sock_flag(sk, SOCK_DEAD))
					sk->sk_state_change(sk);
			}
		}

		/* process RRQ */
		if ((esp_hdr->flags & ESP_FLAG_RRQ) && !(sk->sk_shutdown & SEND_SHUTDOWN)) {
			esp_process_rrq(sk, skb);
			goto out_discard;
		}

		/* check for unwanted data */
		if ((skb->len > 0) && (sk->sk_shutdown & RCV_SHUTDOWN)) {
			esp_send_rst(skb);
			goto out_discard;
		}
		/* check for out-of-sequence data */
		if (seq_after(ESP_SKB_CB(skb)->pkt_seq, esp_sk->recv_seq)) {
#if ESP_USE_OOS_QUEUE
			if (esp_enqueue_oos(sk, skb))
				goto out_queued;
#endif
			goto out_discard;
		}
		/* check for FIN */
		if (esp_hdr->flags & ESP_FLAG_FIN) {
			esp_process_fin(sk, skb);
			goto out_discard;
		}
		/* enqueue in-sequence data */
		switch (esp_enqueue_data(sk, skb)) {
		case 0: /* the packet wasn't enqueued and should be discarded */
			goto out_discard;
		case 1: /* ACK the received and enqueued data packet */
			if (esp_sk->recv_announced) {
				/* we received data so we can defer RRQs */
				//      ESP_RESET_RRQ_TIMER(sk, (unsigned long)HZ);

				if (esp_should_send_ack(esp_sk)) {
					if (!esp_socket_low_rcv_mem(sk))
						esp_enqueue_ack(sk, 0);
					else
						esp_enter_rcv_mem_pressure(sk);
				}
			}

			break;
		case -1:/* the current packet filled a hole and oos packets
       * have been processed - send ACK immediately */
			esp_send_ack(sk);
			break;
		}
		goto out_queued;

	case ESP_CLOSING:
	case ESP_LAST_ACK:
		/* discard out-of-sequence packets and notify the sender */
		if (ESP_SKB_CB(skb)->pkt_seq != esp_sk->recv_seq) {
			esp_send_ack(sk);
			goto out_discard;
		}
		/* check for RST */
		if (esp_hdr->flags & ESP_FLAG_RST) {
			esp_process_rst(sk, skb, -ECONNRESET);
			goto out_discard;
		}
		/* process received ACK */
		if (esp_hdr->flags & ESP_FLAG_ACK) {
			esp_process_ack(sk, skb);
			/* change state when everything has been ACK'ed */
			if (esp_sk->una_seq == esp_sk->send_seq) {
				esp_set_state(sk, ESP_CLOSE);
				if (!sock_flag(sk, SOCK_DEAD)) {
					sk->sk_state_change(sk);
				} else {
					esp_sk_destroy(sk);
				}
			}
		}
		/* ignore anything else */
		goto out_discard;

	default:
		goto out_reset;
	}


out_reset:
	esp_send_rst(skb);

out_discard:
	kfree_skb(skb);

out_queued:
	return err;
}


/*
 * RECVMSG handler
 * Pulls a packet from the receive queue and hands it to the user.
 * Blocks if necessary.
 */
int sock_esp_recvmsg(struct kiocb *iocb, struct socket *sock,
                     struct msghdr *msg, size_t len, int flags)
{
	struct sock *sk = sock->sk;
	struct esp_opt *esp_sk = ESP_SK(sk);
	struct sk_buff *skb = NULL;
	long timeo;
	int target; /* read at least this many bytes */
	int seglen, copied = 0;
	int err = 0;
	ENET_PROFILING_DECLARE(unsigned long long time[2]);

	/* take recvmsg() start time */
	ENET_PROFILING_TIMESTAMP(&time[0]);

	lock_sock(sk);

	if (sk->sk_state == ESP_LISTEN) {
		err = -ENOTCONN;
		goto out;
	}

	/* determine the timeout to use */
	timeo = sock_rcvtimeo(sk, flags & MSG_DONTWAIT);

	/* determine the minimum amount of data to read */
	target = sock_rcvlowat(sk, flags & MSG_WAITALL, len);

	PR_DEBUG("receiving %d...%d bytes on esp socket %p flags %d",
	         target, (unsigned int)len, sk, flags);

	while (copied < len) {
		/* fetch the next data buffer */
		if ((flags & MSG_PEEK) && skb != NULL) {
			/* proceed to the next skb in PEEKing mode */
			skb = skb->next;
			if (skb == (struct sk_buff*)&sk->sk_receive_queue)
				goto out_rcvd;
		} else {
			/* get the first pending skb */
			skb = skb_peek(&sk->sk_receive_queue);
		}

		/* wait for incoming data on blocking sockts */
		if (!skb) {
			/* stop if we read the minimum amount of data needed
			 * and there's nothing in the backlog queue */
			if (copied >= target && !sk->sk_backlog.tail) {
				goto out_rcvd;
			} else {
				/* process the backlog without sleeping */
				release_sock(sk);
				lock_sock(sk);
			}

			/* we need to do the status change checks below as
			 * backlog processing might have changed state.
			 * makes sense? :-)
			 */
			for (;;) {
				/* socket shut down ? */
				if (sk->sk_shutdown & RCV_SHUTDOWN)
					goto out_rcvd;
				/* socket not connected (any more) ? */
				err = sock_error(sk) ? : -ENOTCONN;
				if (sk->sk_state == ESP_CLOSE)
					break;
				/* check for pending signals */
				err = sock_intr_errno(timeo);
				if (signal_pending(current))
					break;
				/* check for nonblocking sockets */
				err = -EAGAIN;
				if (!timeo)
					break;

				err = 0;

				/* check if backlog processing gave us some data */
				if ((skb = skb_peek(&sk->sk_receive_queue)) != NULL) {
					goto skb_found;
				} else {
					/* it seems we really have to sleep... good night, throughput! */
					//printk("reader goes to sleep at %lu.\n", ESP_TIMESTAMP);
					if (sk_wait_data(sk, &timeo)) {
						//printk("reader woke up at %lu.\n", ESP_TIMESTAMP);
						break;
					}
					//printk("reader woke up at %lu in err.\n", ESP_TIMESTAMP);
				}
			}

			if (err)
				goto out_rcvd;

			continue;
		}

skb_found:
		/* determine the amount of data to be copied into
		 * the user buffer */
		seglen = min_t(int, len - copied, skb->len - esp_sk->copy_off);

		PR_DEBUG("copying %d bytes from received packet %d off %d len %d",
		         seglen, ESP_SKB_CB(skb)->pkt_seq,
		         esp_sk->copy_off, skb->len);

		if (!(flags & MSG_TRUNC)) {
			/* copy the data to the user space buffer the
			 * io vector points to */
			err = skb_copy_datagram_iovec(skb,
			                              esp_sk->copy_off, msg->msg_iov, seglen);
			if (err)
				goto out_rcvd;
		}

		/* account for the copied data */
		copied += seglen;

		/* remove the skb if not in PEEKing mode and all of its
		 * data has been copied */
		if (!(flags & MSG_PEEK)) {
			esp_sk->copy_off += seglen;
			if (skb->len == esp_sk->copy_off) {
				/* we copied the whole packet */
				++esp_sk->copy_seq;
				esp_sk->copy_off = 0;
				/* free processed socket buffer */
				__skb_unlink(skb, &sk->sk_receive_queue);
				kfree_skb(skb);

				/* check if the rcv mem pressure has lifted */
				if (esp_sk->rcv_mem_pressure && esp_socket_mem_press_lifted(sk)) {
					esp_lift_rcv_mem_pressure(sk);
					esp_enqueue_ack(sk, esp_sk->need_rtx ? ESP_FLAG_RRQ : 0);
					esp_sk->need_rtx = 0;
				}
			}
		}
	}

out_rcvd:
	PR_DEBUG("received %d bytes on esp socket %p err %d (rq %d mem %d)",
	         copied, sk, err, sk->sk_receive_queue.qlen,
	         atomic_read(&sk->sk_rmem_alloc));

	if (copied > 0)
		err = copied;

out:
	release_sock(sk);

	/* take recvmsg() end time */
	ENET_PROFILING_TIMESTAMP(&time[1]);
	PR_PROFILING("RECVMSG %p %d bytes %llu - %llu (%llu)", sk, copied,
	             time[0], time[1], time[1] - time[0]);

	return err;
}
